import ctypes
import win32api
import win32event
import win32process
import win32security
from win32com.shell import shell as win_shell

user32 = ctypes.windll.user32
advapi32 = ctypes.windll.advapi32

from ctypes.wintypes import HANDLE, DWORD, WORD, LPSTR, BYTE
PROCESS_QUERY_INFORMATION = 0x0400
TOKEN_ADJUST_SESSIONID = 0x0100
INVALID_HANDLE_VALUE = -1
SHGFP_TYPE_CURRENT = 0
CSIDL_PERSONAL = 0x5
NULL = None
TRUE = 1
FALSE = 0

ENABLED_PRIVILEGES = {}

__all__ = ['get_admin_token', 'get_user_token',
           'HumbledProcess', 'set_privilege',
           'get_user_directory']


# typedef struct _PROCESS_INFORMATION {
#    HANDLE hProcess;
#    HANDLE hThread;
#    DWORD dwProcessId;
#    DWORD dwThreadId;
# } PROCESS_INFORMATION, *PPROCESS_INFORMATION, *LPPROCESS_INFORMATION;
class PROCESS_INFORMATION(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ('hProcess', HANDLE),
        ('hThread', HANDLE),
        ('dwProcessId', DWORD),
        ('dwThreadId', DWORD)
        ]


# typedef struct _STARTUPINFO {
#    DWORD   cb;
#    LPSTR   lpReserved;
#    LPSTR   lpDesktop;
#    LPSTR   lpTitle;
#    DWORD   dwX;
#    DWORD   dwY;
#    DWORD   dwXSize;
#    DWORD   dwYSize;
#    DWORD   dwXCountChars;
#    DWORD   dwYCountChars;
#    DWORD   dwFillAttribute;
#    DWORD   dwFlags;
#    WORD    wShowWindow;
#    WORD    cbReserved2;
#    LPBYTE  lpReserved2;
#    HANDLE  hStdInput;
#    HANDLE  hStdOutput;
#    HANDLE  hStdError;
# } STARTUPINFO, *LPSTARTUPINFO;
class STARTUPINFO(ctypes.Structure):
    _pack_ = 1
    _fields_ = [
        ('cb', DWORD),
        ('lpReserved', LPSTR),
        ('lpDesktop', LPSTR),
        ('lpTitle', LPSTR),
        ('dwX', DWORD),
        ('dwY', DWORD),
        ('dwXSize', DWORD),
        ('dwYSize', DWORD),
        ('dwXCountChars', DWORD),
        ('dwYCountChars', DWORD),
        ('dwFillAttribute', DWORD),
        ('dwFlags', DWORD),
        ('wShowWindow', WORD),
        ('cbReserved2', WORD),
        ('lpReserved2', DWORD),
        ('hStdInput', DWORD),
        ('hStdOutput', DWORD),
        ('hStdError', DWORD)
        ]


def get_admin_token(access=win32security.TOKEN_QUERY):
    return win32security.OpenProcessToken(win32process.GetCurrentProcess(), access)


def set_privilege(privilege, state=True):
    if state == ENABLED_PRIVILEGES.get(privilege, False):
        return
    parent_token = get_admin_token(win32security.TOKEN_ADJUST_PRIVILEGES|
                                   win32security.TOKEN_QUERY)
    privilege_luid = win32security.LookupPrivilegeValue(NULL, privilege)
    if state:
        token_privilege = [(privilege_luid, win32security.SE_PRIVILEGE_ENABLED)]
    else:
        token_privilege = [(privilege_luid, 0)]
    try:
        win32security.AdjustTokenPrivileges(parent_token, FALSE, token_privilege)
    finally:
        win32api.CloseHandle(parent_token)
    ENABLED_PRIVILEGES[privilege] = state


def get_user_token():
    set_privilege(win32security.SE_INCREASE_QUOTA_NAME)
    shell_hwnd = user32.GetShellWindow()
    if not shell_hwnd:
        errno = win32api.GetLastError()
        raise WindowsError(errno, win32api.FormatMessage(errno))
    thread_id, process_id = win32process.GetWindowThreadProcessId(shell_hwnd)
    shell_process = win32api.OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, process_id)
    shell_token = win32security.OpenProcessToken(shell_process, win32security.TOKEN_DUPLICATE)
    token_rights = (win32security.TOKEN_QUERY|win32security.TOKEN_ASSIGN_PRIMARY|
                    win32security.TOKEN_DUPLICATE|win32security.TOKEN_ADJUST_DEFAULT|
                    TOKEN_ADJUST_SESSIONID)
    result = win32security.DuplicateTokenEx(shell_token, win32security.SecurityImpersonation,
                                            token_rights, win32security.TokenPrimary)
    win32api.CloseHandle(shell_token)
    return result


def generate_startupinfo():
    info = STARTUPINFO()
    info.cb = ctypes.sizeof(STARTUPINFO)
    info.lpReserved = NULL
    info.lpDesktop = NULL
    info.lpTitle = NULL
    info.dwFlags = 0
    info.cbReserved2 = 0
    info.lpReserved2 = 0
    info.hStdInput = win32api.GetStdHandle(win32api.STD_INPUT_HANDLE)
    info.hStdOutput = win32api.GetStdHandle(win32api.STD_OUTPUT_HANDLE)
    info.hStdError = win32api.GetStdHandle(win32api.STD_ERROR_HANDLE)
    return info


class HumbledProcess:
    def __init__(self, args, executable=NULL, cwd=NULL, creationflags=0):
        if not isinstance(args, unicode):
            raise TypeError('args must be unicode.')
        if not isinstance(cwd, unicode):
            raise TypeError('cwd must be unicode.')
        if not isinstance(executable, unicode):
            raise TypeError('executable must be unicode.')
        self.info = generate_startupinfo()
        self.data = PROCESS_INFORMATION()
        self.data.hProcess = INVALID_HANDLE_VALUE
        self.data.hThread = INVALID_HANDLE_VALUE
        self.data.dwProcessId = 0
        self.data.dwThreadId = 0
        self.args = args
        self.exe = executable
        self.cwd = cwd
        self.flags = creationflags
    
    def start(self):
        token = get_user_token()
        success = advapi32.CreateProcessWithTokenW(token.handle, 0, self.exe,
                                                   self.args, self.flags, NULL,
                                                   self.cwd, ctypes.byref(self.info),
                                                   ctypes.byref(self.data))
        win32api.CloseHandle(token)
        del token
        if success == 0:
            errno = win32api.GetLastError()
            raise WindowsError(errno, win32api.FormatMessage(errno))
    
    def get_pid(self):
        return self.data.dwProcessId
    
    def poll(self):
        if self.data.hProcess != INVALID_HANDLE_VALUE:
            state = win32process.GetExitCodeProcess(self.data.hProcess)
            return state if state != 259 else None
    
    def wait(self, timeout=-1):
        if self.data.hProcess != INVALID_HANDLE_VALUE:
            success = win32event.WaitForSingleObject(self.data.hProcess, timeout)
            if success == win32event.WAIT_OBJECT_0:
                return self.poll()


def get_user_directory(uac_enabled=False):
    token = get_user_token() if uac_enabled else None
    # The CSIDL value refers to the type of path that will be returned.
    # SHGFP_TYPE_CURRENT means that the real path that is currently active
    # is returned and not the default one. (Users may change some of these.)
    result = win_shell.SHGetFolderPath(0, CSIDL_PERSONAL, token, SHGFP_TYPE_CURRENT)
    if uac_enabled:
        CloseHandle(token)
        del token
    return result


if __name__ == '__main__':
    try:
        p = HumbledProcess(u'c:\\windows\\notepad.exe')
        print p.poll()
        p.start()
        print p.wait()
    except Exception, error:
        print error, type(error)
